#!/usr/bin/python
# -*- coding:utf-8 -*-
#
#   Copyright 2008-2009 Le Coz Florent <louizatakk@fedoraproject.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the WTF Public License as published by
#   Sam Hocevar.
#
#   DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#   Version 2, December 2004
#   DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
#   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
#   0. You just DO WHAT THE FUCK YOU WANT TO.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import os
import sys
import xmpp
import select
import pygame as pg
from time import time
from minichat import Chaton
from ownevents import *
from pgu import gui, html
from display_card import *
from pygame.locals import *
from connection import Jabber
from thread import start_new_thread

pg.init()

white = (255, 255, 255)
black = (0, 0, 0)
red = (255, 12, 12)
BLOCK_SIZE = 25
BIG_CARD_SIZE = (int(63*3.5), int(88*3.5))
LITTLE_CARD_SIZE = (63, 88)
FONT = "dejavusans" if "dejavusans" in pg.font.get_fonts() else "freemono"
MAX_CARD_IN_HAND = 11

def load_png(name):
    '''load an image (png OR jpg) and return a surface
    containing this image, and the image\'s rect'''
    fullname = os.path.join('./', name)
    try:
        image = pg.image.load(fullname)
        if image.get_alpha() is None:
            image = image.convert()
        else:
            image = image.convert_alpha()
    except pg.error, message:
        print "Can't load image:", fullname
        raise SystemExit, message
    return (image, image.get_rect())

def find_block(pos):
    '''find the block containing pos'''
    # TODO : remove this shit
    (x, y) = pos
    if x % BLOCK_SIZE < BLOCK_SIZE/2:
        x = x - (x % BLOCK_SIZE)
    else:
        x = x - (x % BLOCK_SIZE) + BLOCK_SIZE
    if y % BLOCK_SIZE < BLOCK_SIZE/2:
        y = y - (y % BLOCK_SIZE)
    else:
        y =  y - (y % BLOCK_SIZE) + BLOCK_SIZE
    return (x, y)

def get_infos(list, name):
    if not list:
        return None
    for card in list:
        if card.dicinfo['name'] == name:
            return card

def nice_resize(img, size):
    try:
        return pg.transform.smoothscale(img, size)
    except:
        return pg.transform.scale(img, size)

def text_to_list2(list):
    max = 42
    li = (list.split('@'))
    new_list = []
    for elem in li:
        max -= elem.count('m')
        max -= elem.count('M')
        max -= elem.count('W')
        max -= elem.count('w')
        while len(elem) > max:
            new_list.append(elem[:find_last_space(elem, max)])
            elem = elem[find_last_space(elem, max)+1:]
        new_list.append(elem)
    return new_list

def create_image(name, info, game):
    '''return the little image representing the card,
    the one moved all over the board'''
    color_dict = {'B': 'black',
                  'U': 'blue',
                  'R': 'red',
                  'G': 'green',
                  'W': 'white',
                  'Gld': 'yellow',
                  'Lnd': 'brown',
                  'Art': 'grey'}
    dic = info.dicinfo
    color = color_dict[dic['color']]
    size = (width, height) = LITTLE_CARD_SIZE
    surface = pg.Surface(size)
    if dic['color'] == 'B':
        c_color = pg.Color('white')
    else:
        c_color = pg.Color('black')
    surface.fill(pg.Color(color))
    font = pg.font.SysFont(FONT, 9)
    name_img = font.render(dic['name'].decode('utf-8'), True, c_color, pg.Color(color))
    font = pg.font.SysFont(FONT, 11)
    try:
        pwr_img = font.render(dic['pwr'], True, c_color, pg.Color(color))
    except: pwr_img = None
    surface.blit(name_img, (2, 2))
    if pwr_img is not None:
        surface.blit(pwr_img,
                     (surface.get_size()[0] - (pwr_img.get_size()[0]+3),
                      surface.get_size()[1] - (pwr_img.get_size()[1]+3)))
    icons = []
    # TODO: handle the W/B-type icons
    for char in dic['cost']:
        try:
            icon, icon_rect = load_png('images/'+char+'.png')
        except:
            sys.stderr.write('%s icon is not recognized' % char)
            continue
        icon = nice_resize(icon, (14, 14))
        icons.append(icon)
    i = surface.get_size()[0] - ((len(icons)*14) + 1)
    for icon in icons:
        surface.blit(icon, (i, 14))
        i += 14
    pg.draw.rect(surface, c_color, surface.get_rect(), 1)
    return surface, surface.get_rect()

def create_big_image(name, infos, game): #FIXME: huge duplicated code from create_image...
    '''return the big image, representing the card, displayed when
    the card is selected'''
    color_dict = {'B': 'black',
                  'U': 'blue',
                  'R': 'red',
                  'G': 'green',
                  'W': 'white',
                  'Gld': 'yellow',
                  'Lnd': 'brown',
                  'Art': 'grey'}
    dic = infos.dicinfo
    info = infos.info
    color = color_dict[dic['color']]
    size = (width, height) = BIG_CARD_SIZE
    surface = pg.Surface(size)

    surface.fill(pg.Color(color))
    font = pg.font.SysFont(FONT, 12)
    if dic['color'] == 'B':
        c_color = pg.Color('white')
    else:
        c_color = pg.Color('black')
    name_img = font.render(dic['name'].decode('utf-8'), True, c_color, pg.Color(color))
    type_img = font.render(dic['type'].decode('utf-8'), True, c_color, pg.Color(color))
    font = pg.font.SysFont(FONT, 15)
    try:
        pwr_img = font.render(dic['pwr'], True, c_color, pg.Color(color))
    except: pwr_img = None
    surface.blit(name_img, (5, 2))
    surface.blit(type_img, (5, 90))
    t_list = text_to_list2(info[5])
    font = pg.font.SysFont(FONT, 9)
    i = 130
    for line in  t_list:
        f_line = font.render(line.decode('utf-8'), True, c_color, pg.Color(color))
        surface.blit(f_line, (5, i))
        i += 15
    if pwr_img is not None:
        surface.blit(pwr_img, (surface.get_size()[0] - (pwr_img.get_size()[0]+3), surface.get_size()[1] - (pwr_img.get_size()[1]+3)))
    icons = []
    for char in dic['cost']:
        try: icon, icon_rect = load_png('images/'+char+'.png')
        except: continue
        icon = nice_resize(icon, (19, 19))
        icons.append(icon)
    i = surface.get_size()[0] - ((len(icons)*19) + 1)
    for icon in icons:
        surface.blit(icon, (i, 2))
        i += 19
    pg.draw.rect(surface, c_color, surface.get_rect(), 1)
    return surface, surface.get_rect()

class Card2(pg.sprite.Sprite):
    '''a sprite representing a card.
    image is the mini_image that we move on the board
    big_image is the complete image displayed as the "selected_card" when
    the card is selected'''

    def __init__(self, image, id=None, pos=(0,0), movable=True, where='hand', list=None, game=None):
        pg.sprite.Sprite.__init__(self)
        self.infos = get_infos(list, image)
        if not self.infos and image != 'back':
            sys.stderr.write('Opponent played an unknown card: "%s"\n' % image)
            sys.stderr.write('Please make sure your opponent has the same cards\n'
                              'than you, in the "cards/" directory.\n'
                              'Game can\'t continue: exiting...')
            sys.exit()
        for root, dir, files in os.walk('images'):
            if '.svn' in dir:
                dir.remove('.svn')
            images = files
        # get mini image
        if image+'.jpg' in images:
            self.image, self.rect = load_png('images/'+image+'.jpg')
            self.image = nice_resize(self.image, LITTLE_CARD_SIZE)
            self.rect = self.image.get_rect()
        else:
            self.image, self.rect = create_image(image, self.infos, game)
        # get complete image
        if image+'.jpg' in images:
            big_image, forget = load_png('images/'+image+'.jpg')
            self.big_image = nice_resize(big_image, BIG_CARD_SIZE)
        else:
            self.big_image, forget = create_big_image(image, self.infos, game)
        self.name = image
        self.id = id
        self.movable = movable
        self.rect.center = pos
        self.where = where
        self.tapped = False
        return

    def rotate(self, tap=None):
        '''rotate the card'''
        center = self.rect.center
        if self.tapped == True and tap != True:
            self.image = pg.transform.rotate(self.image, 90)
            self.tapped = False
        elif self.tapped == False and tap != False:
            self.image = pg.transform.rotate(self.image, -90)
            self.tapped = True
        else:
            return
        self.rect = self.image.get_rect()
        if self.tapped == True:
            self.rect.center = (center[0], center[1]+21)
            return False
        else:
            self.rect.center = (center[0], center[1]-21)
            return True

class Stack(Card2):
    '''a Card2, but with a number written on it
    and some methods that can change this number'''
    def __init__(self, image, id=None, pos=(0,0), movable=True, where='hand',
                 list=None, card_number=None, game=None):
        Card2.__init__(self, image, id, pos, movable, where, list, game)
        self.number = card_number
        self.paint_number()

    def draw(self):
        self.number -= 1
        self.paint_number()

    def undraw(self):
        self.number += 1
        self.paint_number()

    def paint_number(self):
        font = pg.font.SysFont(FONT, 9)
        if self.number >= 10:
            f_line = font.render(str(self.number), True, pg.Color('black'), pg.Color('red'))
        else:
            f_line = font.render('0'+str(self.number), True, pg.Color('black'), pg.Color('red'))
        self.image.blit(f_line, (5, 2))

class Selected(pg.sprite.Sprite):
    '''a sprite, never moving, just showing the big_image of a Card2
    whenever a card is selected
    should be instanciated ONCE'''
    def __init__(self):
        pg.sprite.Sprite.__init__(self)
        image, rect = load_png('images/back.jpg')
        self.image = pg.transform.scale(image, BIG_CARD_SIZE)
        self.rect = self.image.get_rect()

class GameZone(pg.sprite.Sprite):
    '''an invisible (visible if needed for debug) gamezone, to handle click and/or drag'n'drop'''
    def __init__(self, pos, size):
        pg.sprite.Sprite.__init__(self)
        self.image = pg.Surface(size)
        self.rect = self.image.get_rect()
        pg.draw.rect(self.image, pg.Color('white'), self.rect, 1)
        self.rect.left = pos[0]
        self.rect.top = pos[1]

class ScoreZone(GameZone):
    '''A zone where the player's score is displayed'''
    def __init__(self, pos, size, life=20):
        GameZone.__init__(self, pos, size)
        self.life = life
        self.size = size
        self.paint_number()

    def paint_number(self):
        font = pg.font.SysFont(FONT, 39)
        if self.life < 10 and self.life >= 0:
            f_line = font.render('0'+str(self.life), True, pg.Color('black'), pg.Color('red'))
        else:
            f_line = font.render(str(self.life), True, pg.Color('black'), pg.Color('red'))
        self.image.blit(f_line, (0, 0))

    def change(self, nb):
        if self.life + nb > 99 or self.life + nb < -9:
            return
        self.life += nb
        self.paint_number()

class Game(object):
    '''the complete Game (should be instanciated ONCE), with the complete board
    handle events, and uses
    a jabber class to send message to the other client'''

    def __init__(self, list, deck, jabber):
        self.deck = deck
        self.list = list
        self.jabber = jabber
        self.screen = pg.display.set_mode()
        self.size = (self.width, self.height) = self.screen.get_size()
        self.arrow_list = []
        self.selected_card = Selected()
        self.selected_card.rect.right = (self.width - 5)
        self.selected_card.rect.top = self.h(100)
        self.zone_hand = GameZone((self.w(100), self.height-self.h(150)), (self.width-self.w(100), self.h(150)))
        self.zone_cim = GameZone((2, self.h(822)), (LITTLE_CARD_SIZE[0]*1.5, LITTLE_CARD_SIZE[1]*1.5))
        self.zone_opcim = GameZone((2, self.h(180)), (LITTLE_CARD_SIZE[0]*1.5, LITTLE_CARD_SIZE[1]*1.5))
        self.zone_stack = GameZone((25, self.height-130), LITTLE_CARD_SIZE)

        self.bg, self.bg_rect = load_png('images/plateaum.png')
        self.bg = pg.transform.scale(self.bg, self.size)
        self.overcard = None
        self.moving = False

        self.last_redraw = 0
        self.start_arrow = False
        self.screen.blit(self.bg, self.bg_rect)

        self.stack = Stack('back', 0, (self.w(60), self.height-self.h(80)), False, None, None, len(self.deck))
        self.opstack = Stack('back', 0, (self.w(60), self.h(80)), False, None, None, 0)

        self.life = ScoreZone((2, self.height/2), (50, 50))
        self.oplife = ScoreZone((2, self.height/2-50), (50, 50))
        self.cards = pg.sprite.OrderedUpdates(self.stack, self.opstack)
        #debug help : put the GameZones in self.update to see them
        self.update = pg.sprite.OrderedUpdates(self.selected_card, self.life, self.oplife)#, self.zone_cim, self.zone_opcim, self.zone_hand)
        self.lastid = 1
        self.chat = Chaton(200, 300, self.screen, self.jabber)
        self.chat.my_loop(None)

    def w(self, width):
        '''use self.width to adjust the coordinate with the
        screen resolution'''
        return int(float(width)*float(self.width)/1900)

    def unw(self, width):
        '''use self.width to adjust the coordinate with the
        screen resolution'''
        return int(float(width)/float(self.width)*1900)

    def h(self, height):
        '''use self.height to adjust the coordinate with the
        screen resolution'''
        return int(float(height)*float(self.height)/1200)

    def unh(self, height):
        '''use self.height to adjust the coordinate with the
        screen resolution'''
        return int(float(height)/float(self.height)*1200)

    def get_card_under_mouse(self, pos):
        '''return the Card2 over which the cursor is'''
        temp = list(self.cards)
        temp.reverse()
        for card in temp:
            if card.rect.collidepoint(pos):
                return card
        return None

    def get_card(self, id):
        '''retun the card with id=id'''
        for card in self.cards:
            if int(card.id) == int(id):
                return card
        return None

    def move_card(self, id, pos):
        '''Move Card2 with id=id to pos position'''
        card = self.get_card(int(id))
        card.rect.center = pos

    def next_id(self):
        '''return the next available card id'''
        self.lastid += 1
        return self.lastid

    def nb_cards_in_hand(self, op=False):
        '''return the number of cards of the player's hand
        if op==True, the number of cards in the opponent's
        hand is returned'''
        if op == True:
            search = 'ophand'
        else:
            search = 'hand'
        nb = 0
        for card in self.cards:
            try:
                if card.where == search:
                    nb += 1
            except:pass
        return nb

    def nb_cards_in_cim(self, op=False):
        '''return the number of cards of the player's cim
        if op==True, the number of cards in the opponent's
        hand is returned'''
        if op == True:
            search = 'opcim'
        else:
            search = 'cim'
        nb = 0
        for card in self.cards:
            try:
                if card.where == search:
                    nb += 1
            except:pass
        return nb

    def update_hand(self, op=False, space=80):
        '''update the player/op\'s hand'''
        if op == True:
            nb = self.nb_cards_in_hand(op=True)
            start = self.size[0]/2 + space - (space/2 * nb) - 100
            for card in self.cards:
                if card.where == 'ophand':
                    card.rect.center = (start, 80)
                    start += space
            return
        nb = self.nb_cards_in_hand()
        start = self.size[0]/2 + space - (space/2 * nb) - 100
        for card in self.cards:
            if card.where == 'hand':
                card.rect.center = (start, self.size[1]-80)
                start += space

    def update_cim(self, op=False):
        '''update the player/op\'s cim'''
        if op == True:
            for card in self.cards:
                if card.where == 'opcim':
                    card.rect.center = self.zone_opcim.rect.center
            return
        for card in self.cards:
            if card.where == 'cim':
                card.rect.center = self.zone_cim.rect.center

    def rotate_cim(self, op, reverse=False):
        '''change the card that is above the cim
        actual method is: find the first (ie: at the botom)
        card which is in cim, and put it on top on the self.cards'''
        # get the topcard of the cim (i.e. the last one)
        where = 'cim' if op == False else 'opcim'
        new = None
        for card in self.cards:
            if card.where == where:
                last = card
        if reverse == False:
            for card in self.cards:
                if card.where == where:
                    new = card
                    break
            if not new:
                return
            new.remove(self.cards)
            new.add(self.cards)
            self.selected_card.image = new.big_image
        else:
            for card in self.cards:
                if card.where == where and card != last:
                    card.remove(self.cards)
                    card.add(self.cards)
            self.selected_card.image = last.big_image

    def draw(self, op=False, id=None):
        '''take the first card in self.deck and put it in hand
        and update the number displayed on the Stack'''
        if not op:
            if len(self.deck) > 0:
                id = self.next_id()
                name = self.deck.pop(0)
                self.cards.add(Card2(name, int(id), (0,0), False, 'hand', self.list, game=self))
                self.update_hand(False)
                self.stack.draw()
                return id
            return None
        else:
            self.cards.add(Card2('back', int(id), (0,0), False, 'ophand', self.list, game=self))
            self.update_hand(op=True)
            self.opstack.draw()
            return None

    def get_free_place(self, type):
        '''return a position on the board where
        there is no card at the moment'''
        height_list = (self.height-self.h(300), self.height-self.h(450), self.height-self.h(580)) if 'Land' in type.split() else (self.height-self.h(440), self.height-self.h(570))
        for height in height_list:
            for width in xrange(self.w(300), self.w(1500), self.w(LITTLE_CARD_SIZE[0]+60)):
                free = True
                for card in self.cards:
                    if card.rect.collidepoint((width, height)):
                        free = False
                if free == True:
                    return (width, height)
        return (width, height)

    def handle_left_click(self, pos):
        self.selected_card.image = self.overcard.big_image
        if self.overcard.movable == False:
            return
        self.overcard.remove(self.cards)
        self.overcard.add(self.cards)
        self.moving = True
        self.overcard.rect.center = pos
        return(True)

    def handle_right_click(self, pos):
        if not self.overcard:
            return False
        if self.overcard == self.stack:
            if self.nb_cards_in_hand() <= MAX_CARD_IN_HAND:
                id = self.draw()
                if id:
                    self.jabber.send_draw(-id)
        elif self.overcard and self.overcard.where == 'hand':
            self.overcard.where = 'game'
            self.overcard.movable = True
            free_pos = self.get_free_place(self.overcard.infos.dicinfo['type'])
            self.move_card(self.overcard.id, free_pos)
            self.update_hand()
            self.jabber.send_edit_card({'name':self.overcard.name,
                                   'x':self.unw(free_pos[0]),
                                    'y':self.unh(self.height-free_pos[1]),
                                   'id':-self.overcard.id,
                                   'where':'game',
                                    'tapped':False})
        elif self.overcard.where == 'game'and self.overcard.movable == True:
            ro = self.overcard.rotate()
            ro = False if ro else True
            self.jabber.send_edit_card({'id':-self.overcard.id,
                                        'tapped':ro})
        return True

    def handle_middle_click(self, pos):
        card = self.get_card_under_mouse(pos)
        if card and card.where == 'game':
            self.start_arrow = card
            return False
        elif card and card.where == 'hand':
            self.stack.undraw()
            self.overcard.remove(self.cards)
            self.deck.append(self.overcard.name)
            self.update_hand()
            self.jabber.send_undraw(-self.overcard.id)
        return True

    def return_card_in_hand(self):
        self.overcard.where = 'hand'
        self.overcard.movable = False
        self.overcard.rotate(False)
        self.update_hand()
        self.jabber.send_edit_card({'id':-self.overcard.id,
                                   'where':'ophand',
                                    'name':'back',
                                    'tapped': False})

    def put_card_in_cim(self):
        self.overcard.where = 'cim'
        self.overcard.movable = True
        self.overcard.rotate(False)
        self.jabber.send_edit_card({'id':-self.overcard.id,
                                    'where':'opcim',
                                    'tapped': False})
        self.update_cim()

    def return_card_in_stack(self):
        self.stack.undraw()
        self.overcard.remove(self.cards)
        self.deck.append(self.overcard.name)
        self.jabber.send_undraw(-self.overcard.id)

    def edit(self, event):
        id = int(event.id)
        oldcard = self.get_card(id)
        if not oldcard:
            print 'Trying to edit card #%s, but doesn\'t seem to exist...' % id
            return
        if event.where and event.where != 'None':
            oldcard.where = event.where
        if event.x and event.x != 'None' and event.y and event.y != 'None':
            oldcard.rect.center = (self.w(int(event.x)), self.h(int(event.y)))
        if event.tapped and event.tapped != 'None':
            if event.tapped == True or event.tapped == 'True':
                oldcard.rotate(True)
            elif event.tapped == False or event.tapped == 'False':
                oldcard.rotate(False)
        if event.name and event.name != 'None':
            newcard = Card2(str(event.name), 0, (0,0), False, 'hand', self.list, game=self)
            if not newcard:
                print 'Error : could not create a card named %s' % event.name
                return
            oldcard.image = newcard.image
            oldcard.big_image = newcard.big_image
        oldcard.remove(self.cards)
        oldcard.add(self.cards)
        self.update_hand(True)
        self.update_hand(False)
        self.update_cim(True)
        self.update_cim(False)

    def main_loop(self):
        ''' main loop checking for user OR jabber event.
        Redraw the board if needed'''
        event = pg.event.wait()
        try:self.chat.my_loop(event)
        except:pass
        redraw = False
        if event.type == QUIT:
            sys.exit()
        if event.type == EVENTINIT:
            self.jabber.connected = True
            self.opstack.number = int(event.deck_size)
            self.opstack.paint_number()
            redraw = True
        if self.jabber.connected == False:
            return
        if pg.event.peek(6):
            pg.event.clear(4)
        pos = (x, y) = pg.mouse.get_pos()
        if event.type == EVENTMOVE:
            self.move_card(event.id, (int(event.x), self.height - int(event.y)))
            redraw = True
        elif event.type == EVENTDRAW:
            self.draw(op=True, id=event.id)
            redraw = True
        elif event.type == EVENTUNDRAW:
            self.opstack.undraw()
            card = self.get_card(int(event.id))
            self.cards.remove(card)
            self.update_hand(True)
            redraw = True
        elif event.type == EVENTEDIT:
            self.edit(event)
            redraw = True
        elif event.type == EVENTMSG:
            self.chat.write_line(event.message.encode('latin-1'), '< ')
            us = self.chat.update(self.chat.screen)
            pg.display.update(us)
            redraw = True
        elif event.type == EVENTLIFE:
            self.oplife.change(int(event.nb))
        elif event.type == EVENTUNTAP_ALL:
            for card in self.cards:
                if card.where == 'game' and card.movable == False:
                    card.rotate(False)
            redraw = True
        elif self.moving == False:
            self.overcard = self.get_card_under_mouse(pg.mouse.get_pos())
            if self.overcard and self.overcard != self.stack and self.overcard != self.opstack:
                if self.selected_card.image != self.overcard.big_image:
                    self.selected_card.image = self.overcard.big_image
                    redraw = True

        ## on clic down
        if event.type == 5:
            if self.overcard != None:
                if event.button == 1:
                    redraw = self.handle_left_click(pos)
                elif event.button == 3:
                    redraw = self.handle_right_click(pos)
                if event.button == 2:
                    redraw = self.handle_middle_click(pos)

        ## mouse move
        elif event.type == 4 and event.buttons == (1, 0, 0) and time() - self.last_redraw > 0.02:
            if self.moving == True and self.overcard != None:
                self.overcard.rect.center = pos
                if self.overcard.rect.center[0] > self.width-240:
                    self.overcard.rect.center = (self.width-240, pos[1])
                redraw = True
                self.last_redraw = time()

        ## clic up or scroll
        elif event.type == 6:
            redraw = True
            pos = (x, y) = pg.mouse.get_pos()
            if self.overcard != None and self.moving == True:
                # We drop the card, so we clear
                # all the "mouse moving" events
                pg.event.clear(4)
                self.moving = False
                if self.zone_hand.rect.collidepoint(pos):
                    self.return_card_in_hand()
                elif self.zone_cim.rect.collidepoint(pos):
                    self.put_card_in_cim()
                elif self.zone_stack.rect.collidepoint(pos):
                    self.return_card_in_stack()
                else:
                    self.overcard.where = 'game'
                    if self.jabber:
                        self.jabber.send_edit_card({'id':-int(self.overcard.id),
                                                    'x':self.unw(pos[0]),
                                                    'y':self.unh(self.height-pos[1]),
                                                    'where':'game',
                                                    })
            if event.button == 2 and self.start_arrow:
                card = self.get_card_under_mouse(pos)
                if card:
                    self.arrow_list.append((self.start_arrow, card))
                    self.start_arrow = None
            elif event.button == 4 and self.zone_cim.rect.collidepoint(pos):
                self.rotate_cim(False)
            elif event.button == 4 and self.zone_opcim.rect.collidepoint(pos):
                self.rotate_cim(True)
            elif event.button == 5 and self.zone_cim.rect.collidepoint(pos):
                self.rotate_cim(False, True)
            elif event.button == 5 and self.zone_opcim.rect.collidepoint(pos):
                self.rotate_cim(True, True)

        ## keyboard
        elif event.type == 2:
            if event.key == 108 and (event.mod == 4160 or event.mod == 64): ## ctrl+L
                self.arrow_list = []
                self.screen.blit(self.bg, self.bg_rect)
            elif event.key == 117 and (event.mod == 4160 or event.mod == 64): ## ctrl+u
                for card in self.cards:
                    if card.where == 'game' and card.movable == True:
                        card.rotate(False)
                self.jabber.send_untap_all()
            elif event.key == 100 and (event.mod == 4160 or event.mod == 64): ## ctrl+d
                if self.nb_cards_in_hand() <= MAX_CARD_IN_HAND:
                    id = self.draw()
                    if id:
                        self.jabber.send_draw(-id)
            elif event.key == pg.K_PAGEUP:
                self.jabber.send_life_change(1)
                self.life.change(1)
            elif event.key == pg.K_PAGEDOWN:
                self.jabber.send_life_change(-1)
                self.life.change(-1)
            redraw = True
        ## redraw everything
        if redraw:
            if len(self.arrow_list) > 0 and self.moving == False:
                self.screen.blit(self.bg, self.bg_rect)
                for arrow in self.arrow_list:
                    pg.draw.line(self.screen, red, arrow[0].rect.center, arrow[1].rect.center, 7)
            self.cards.clear(self.screen, self.bg)
            changes = self.update.draw(self.screen)
            pg.display.update(changes)
            changes = self.cards.draw(self.screen)
            pg.display.update(changes)
            pg.display.flip()
        return

if __name__ == '__main__':
    game = Game()
    while True:
        game.main_loop()
